/**
 * data-structure-typed
 *
 * @author Tyler Zeng
 * @copyright Copyright (c) 2022 Tyler Zeng <zrwusa@gmail.com>
 * @license MIT License
 */
import { IterableWithSizeOrLength, IterateDirection } from "../../types";
/**
 * Deque can provide random access with O(1) time complexity
 * Deque is usually more compact and efficient in memory usage because it does not require additional space to store pointers.
 * Deque may experience performance jitter, but DoublyLinkedList will not
 * Deque is implemented using a dynamic array. Inserting or deleting beyond both ends of the array may require moving elements or reallocating space.
 */
export declare class DequeIterator<E> {
    iterateDirection: IterateDirection;
    index: number;
    readonly deque: Deque<E>;
    constructor(index: number, deque: Deque<E>, iterateDirection?: IterateDirection);
    get current(): E;
    set current(newElement: E);
    isAccessible(): boolean;
    prev(): DequeIterator<E>;
    next(): DequeIterator<E>;
    clone(): DequeIterator<E>;
}
export declare class Deque<E> {
    protected _bucketFirst: number;
    protected _firstInBucket: number;
    protected _bucketLast: number;
    protected _lastInBucket: number;
    protected _bucketCount: number;
    protected readonly _bucketSize: number;
    constructor(elements?: IterableWithSizeOrLength<E>, bucketSize?: number);
    protected _buckets: E[][];
    get buckets(): E[][];
    protected _size: number;
    get size(): number;
    get first(): E | undefined;
    get last(): E | undefined;
    /**
     * Time Complexity: Amortized O(1) - Generally constant time, but resizing when the deque is full leads to O(n).
     * Space Complexity: O(n) - In worst case, resizing doubles the array size.
     */
    empty(): boolean;
    /**
     * Time Complexity: O(1) - Removes the last element.
     * Space Complexity: O(1) - Operates in-place.
     */
    isEmpty(): boolean;
    /**
     * Time Complexity: Amortized O(1) - Similar to push, resizing leads to O(n).
     * Space Complexity: O(n) - Due to potential resizing.
     */
    /**
     * Time Complexity: O(1)
     * Space Complexity: O(n) - In worst case, resizing doubles the array size.
     *
     * The addLast function adds an element to the end of an array.
     * @param {E} element - The element parameter represents the element that you want to add to the end of the
     * data structure.
     */
    addLast(element: E): void;
    /**
     * Time Complexity: O(1) - Removes the first element.
     * Space Complexity: O(1) - In-place operation.
     */
    /**
     * Time Complexity: O(1) - Removes the last element.
     * Space Complexity: O(1) - Operates in-place.
     *
     * The function "popLast" removes and returns the last element of an array.
     * @returns The last element of the array is being returned.
     */
    popLast(): E | undefined;
    /**
     * Time Complexity: O(1).
     * Space Complexity: O(n) - Due to potential resizing.
     *
     * The "addFirst" function adds an element to the beginning of an array.
     * @param {E} element - The parameter "element" represents the element that you want to add to the
     * beginning of the data structure.
     */
    addFirst(element: E): void;
    /**
     * Time Complexity: O(1) - Removes the first element.
     * Space Complexity: O(1) - In-place operation.
     *
     * The function "popFirst" removes and returns the first element of an array.
     * @returns The method `popFirst()` is returning the first element of the array after removing it
     * from the beginning. If the array is empty, it will return `undefined`.
     */
    popFirst(): E | undefined;
    clear(): void;
    begin(): DequeIterator<E>;
    end(): DequeIterator<E>;
    reverseBegin(): DequeIterator<E>;
    reverseEnd(): DequeIterator<E>;
    push(element: E): number;
    pop(): E | undefined;
    unshift(element: E): number;
    shift(): E | undefined;
    getAt(pos: number): E;
    setAt(pos: number, element: E): void;
    insertAt(pos: number, element: E, num?: number): number;
    cut(pos: number): number;
    deleteAt(pos: number): number;
    delete(element: E): number;
    deleteByIterator(iter: DequeIterator<E>): DequeIterator<E>;
    findIterator(element: E): DequeIterator<E>;
    reverse(): this;
    unique(): number;
    sort(comparator?: (x: E, y: E) => number): this;
    shrinkToFit(): void;
    forEach(callback: (element: E, index: number, deque: Deque<E>) => void): void;
    find(callback: (element: E, index: number, deque: Deque<E>) => boolean): E | undefined;
    toArray(): E[];
    map<T>(callback: (element: E, index: number, deque: Deque<E>) => T): Deque<T>;
    filter(predicate: (element: E, index: number, deque: Deque<E>) => boolean): Deque<E>;
    reduce<T>(callback: (accumulator: T, element: E, index: number, deque: Deque<E>) => T, initialValue: T): T;
    indexOf(element: E): number;
    [Symbol.iterator](): Generator<E, void, unknown>;
    protected _reallocate(needBucketNum?: number): void;
    protected _getBucketAndPosition(pos: number): {
        bucketIndex: number;
        indexInBucket: number;
    };
}
export declare class ObjectDeque<E = number> {
    constructor(capacity?: number);
    protected _nodes: {
        [key: number]: E;
    };
    get nodes(): {
        [p: number]: E;
    };
    protected _capacity: number;
    get capacity(): number;
    protected _first: number;
    get first(): number;
    protected _last: number;
    get last(): number;
    protected _size: number;
    get size(): number;
    /**
     * Time Complexity: O(1)
     * Space Complexity: O(1)
     */
    /**
     * Time Complexity: O(1)
     * Space Complexity: O(1)
     *
     * The "addFirst" function adds an element to the beginning of an array-like data structure.
     * @param {E} element - The `element` parameter represents the element that you want to add to the beginning of the data
     * structure.
     */
    addFirst(element: E): void;
    /**
     * Time Complexity: O(1)
     * Space Complexity: O(1)
     */
    /**
     * Time Complexity: O(1)
     * Space Complexity: O(1)
     *
     * The addLast function adds an element to the end of an array-like data structure.
     * @param {E} element - The `element` parameter represents the element that you want to add to the end of the data structure.
     */
    addLast(element: E): void;
    /**
     * Time Complexity: O(1)
     * Space Complexity: O(1)
     */
    /**
     * Time Complexity: O(1)
     * Space Complexity: O(1)
     *
     * The function `popFirst()` removes and returns the first element in a data structure.
     * @returns The element of the first element in the data structure.
     */
    popFirst(): E | undefined;
    /**
     * Time Complexity: O(1)
     * Space Complexity: O(1)
     */
    /**
     * Time Complexity: O(1)
     * Space Complexity: O(1)
     *
     * The `getFirst` function returns the first element in an array-like data structure if it exists.
     * @returns The element at the first position of the `_nodes` array.
     */
    getFirst(): E | undefined;
    /**
     * Time Complexity: O(1)
     * Space Complexity: O(1)
     */
    /**
     * Time Complexity: O(1)
     * Space Complexity: O(1)
     *
     * The `popLast()` function removes and returns the last element in a data structure.
     * @returns The element that was removed from the data structure.
     */
    popLast(): E | undefined;
    /**
     * Time Complexity: O(1)
     * Space Complexity: O(1)
     */
    /**
     * Time Complexity: O(1)
     * Space Complexity: O(1)
     *
     * The `getLast()` function returns the last element in an array-like data structure.
     * @returns The last element in the array "_nodes" is being returned.
     */
    getLast(): E | undefined;
    /**
     * Time Complexity: O(1)
     * Space Complexity: O(1)
     */
    /**
     * Time Complexity: O(1)
     * Space Complexity: O(1)
     *
     * The get function returns the element at the specified index in an array-like data structure.
     * @param {number} index - The index parameter is a number that represents the position of the element you want to
     * retrieve from the array.
     * @returns The element at the specified index in the `_nodes` array is being returned. If there is no element at that
     * index, `undefined` is returned.
     */
    get(index: number): NonNullable<E> | undefined;
    /**
     * The function checks if the size of a data structure is less than or equal to zero.
     * @returns The method is returning a boolean element indicating whether the size of the object is less than or equal to 0.
     */
    isEmpty(): boolean;
}
